;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "dune" "20260410.1721"
  "Integration with the dune build system."
  ()
  :url "https://github.com/ocaml/dune"
  :commit "df26745e52be99ecdb2ff994feab1eacd858b226"
  :revdesc "df26745e52be")
